#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
nti.schema.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import zope.i18nmessageid

__docformat__ = "restructuredtext en"

MessageFactory = zope.i18nmessageid.MessageFactory('nti.dataserver')
